/*
 * Copyright (c) 2015 N. Billard.
 */
package com.massa.mrules.demos.simplexml;

import static org.junit.Assert.assertEquals;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.mrules.context.compile.CompilationLevel;
import com.massa.mrules.context.compile.MInOutCompilationContext;
import com.massa.mrules.context.execute.MInOutExecutionContext;
import com.massa.mrules.demos.simplexml.model.ClientType;
import com.massa.mrules.demos.simplexml.model.Command;
import com.massa.mrules.demos.simplexml.model.CommandConstraints;
import com.massa.mrules.demos.simplexml.model.ProductType;
import com.massa.mrules.factory.ser.SerializationFactory;
import com.massa.mrules.factory.xml.SimpleXMLFactory;
import com.massa.mrules.set.IMruleExecutionSet;
import com.massa.util.convert.ConvertUtils;

@RunWith(Parameterized.class)
public class SimpleXmlTest {
	private static final Log LOG = LogFactory.getLog(SimpleXmlTest.class.getName());

	final CompilationLevel level;

	public SimpleXmlTest(final CompilationLevel level) {
		this.level = level;
	}

	@Parameters
	public static Collection<Object[]> configs() {
		return Arrays.asList(new Object[][] {
				{CompilationLevel.STANDARD},
				{CompilationLevel.LIGHT}
		});
	}

	private static final String XML = "com/massa/mrules/demos/simplexml/rules.xml";

	/*
	Rules details (in this order) :
	 - If France and book since 2016-01-01 : VAT 5.5% and 3% max discount.
	 - If France and book : VAT 5.5% and 5% max discount.
	 - If Association : VAT 5.5%.
	 - If book : VAT 10%.
	 - If Food : VAT 5.5% and 0% max discount.
	 - If Society and electronics : VAT 10% and 15% max discount.
	 - If Person and type in house or music : VAT 10% and 25% max discount.
	 - In any other case : VAT 20%.
	 */

	String[][] inputs = new String[][]{
			new String[]{ "SOCIETY", "FRANCE", "FOOD", "2016-01-01" },
			new String[]{ "ASSOCIATION", "SPAIN", "FOOD", "2015-11-01" },
			new String[]{ "SOCIETY", "GERMANY", "ELECTRONIC", "2015-06-01" },
			new String[]{ "ASSOCIATION", "FRANCE", "BOOK", "2016-02-01" },
			new String[]{ "PERSON", "FRANCE", "BOOK", "2015-08-01" },
			new String[]{ "SOCIETY", "ITALY", "MUSIC", "2016-02-01" },
			new String[]{ "PERSON", "ITALY", "MUSIC", "2015-01-01" },
			new String[]{ "SOCIETY", "FRANCE", "ELECTRONIC", "2015-03-01" },
			new String[]{ "ASSOCIATION", "FRANCE", "ELECTRONIC", "2016-01-01" },
			new String[]{ "PERSON", "FRANCE", "HOUSE", "2015-12-15" },
	};

	String[][] outputs = new String[][]{
			new String[]{ "VAT_55", "5.5", "0" },
			new String[]{ "VAT_55", "5.5", null },
			new String[]{ "VAT_10", "10", "15" },
			new String[]{ "VAT_55", "5.5", "3" },
			new String[]{ "VAT_55", "5.5", "5" },
			new String[]{ "VAT_20", "20", null },
			new String[]{ "VAT_10", "10", "25" },
			new String[]{ "VAT_10", "10", "15" },
			new String[]{ "VAT_55", "5.5", null },
			new String[]{ "VAT_10", "10", "25" },
	};

	@Test
	public void simpleXmlDemo() throws Exception {
		SimpleXMLFactory factory = new SimpleXMLFactory(
				new MInOutCompilationContext(this.level, Command.class, CommandConstraints.class));

		InputStream xml = ClassLoader.getSystemResourceAsStream(XML);
		IMruleExecutionSet eng = factory.read(xml);

		Command command = new Command();
		CommandConstraints result = new CommandConstraints();
		MInOutExecutionContext<Command, CommandConstraints> ctxt = new MInOutExecutionContext<Command, CommandConstraints>(
				eng, command, result);
		ctxt.execute();

		for (int idx = 0 ; idx < this.inputs.length ; idx++) {
			this.reset(command, result);
			command.setClientType(ConvertUtils.getDefaultInstance().cast(ClientType.class, this.inputs[idx][0]));
			command.setRegion(this.inputs[idx][1]);
			command.setProductType(ConvertUtils.getDefaultInstance().cast(ProductType.class, this.inputs[idx][2]));
			command.setCommandDate(ConvertUtils.getDefaultInstance().cast(Date.class, this.inputs[idx][3]));

			ctxt.execute();

			assertEquals("index: " + idx, this.outputs[idx][0], result.getVat().getLabel());
			assertEquals("index: " + idx, ConvertUtils.getDefaultInstance().cast(BigDecimal.class, this.outputs[idx][1]), result.getVat().getRate());
			assertEquals("index: " + idx, ConvertUtils.getDefaultInstance().cast(BigDecimal.class, this.outputs[idx][2]), result.getMaxDiscount());
		}

		//Time check
		LOG.info("Beginning performance test.");
		long t1 = System.currentTimeMillis();
		for (int i = 0 ; i < 10000 ; i++) {
			for (int idx = 0 ; idx < this.inputs.length ; idx++) {
				this.reset(command, result);
				command.setClientType(ConvertUtils.getDefaultInstance().cast(ClientType.class, this.inputs[idx][0]));
				command.setRegion(this.inputs[idx][1]);
				command.setProductType(ConvertUtils.getDefaultInstance().cast(ProductType.class, this.inputs[idx][2]));
				command.setCommandDate(ConvertUtils.getDefaultInstance().cast(Date.class, this.inputs[idx][3]));

				ctxt.execute();
			}
		}
		LOG.info("Time for 10000 tests : " + (System.currentTimeMillis() - t1));

		//Read / Write tests.
		LOG.info("Human readable version:\n" + eng.toString());

		StringWriter w = new StringWriter();
		factory.write(w, eng);
		LOG.info(w.toString());

		ByteArrayOutputStream o = new ByteArrayOutputStream();
		SerializationFactory sfactory = new SerializationFactory(this.level);
		sfactory.write(o, eng);

		Assert.assertEquals(eng, eng.clone());
		Assert.assertEquals(eng, factory.read(new StringReader(w.toString())));
		Assert.assertEquals(eng, sfactory.read(new ByteArrayInputStream(o.toByteArray())));
	}

	private void reset(final Command command, final CommandConstraints result) {
		command.setCommandDate(null);
		command.setProductType(null);
		command.setRegion(null);
		command.setClientType(null);
		result.setMaxDiscount(null);
		result.setVat(null);
	}
}
